import org.w3c.tidy.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

/*
By: 	Kinga Laura Dobolyi
	University of Virginia, Department of Computer Science
	dobolyi@virginia.edu
	http://www.cs.virginia.edu/~kld5r/
	
	Appears in "Modeling Consumer-Perceived Web Application Fault Severities for Testing" by Kinga Dobolyi 
	and Westley Weimer, International Symposium on Software Testing and Analysis. July 2010.
	http://www.cs.virginia.edu/~kld5r/issta2010.pdf
	
	
README:
	This class is meant to be used in conjuction with the shell script paper3Web.py
	Requirements (for running in UNIX environment):
		- paper3Web.sh (the shell script that processes all of your input HTML files)
		- AutomaticSeverity.java (this file)
		- Tidy.jar (used in this class, needs to be on your classpath)
		- parseBothPaper3.py (creates a mapping between two HTML files)
		- html5lib (python library)
		- xml (python library)
		- file.txt (lists all the input file pairs to be processed)
		- faults/... (the folder of the input file pairs)
		
	All of the necessary files and libraries come shipped with the tool. If you unzip AutomaticSeverity.zip and therefore	
	if you have all of these files in the same directory, you can run the tool as:
		%sh paper3Web.sh
		
	paper3Web.sh expects a file called file.txt that contains the names of all the input files
	you with to pass to this tool. The input files must be in a subfolder called "faults" and
	must end in ..._correctHTML.html and ..._incorrectHTML.html in order to be processed. 
	See the example correct and incorrect file provided (in the faults folder).
	
OUTPUT:
	This file will output results to a file called automatic.csv listing the severity rating
	under the column SCORE, followed by feature values.
	
	The rest of the output is generated by the parsing libraries and can be ignored. Note that
	the parsing libraries may take up to 30 seconds to parse files, and will time out afterwards.
*/

class AutomaticSeverity 
{
	//used for operating system calls
	static Runtime run = Runtime.getRuntime();
	static Process pr;
	static BufferedReader buf;
	static String line = null;
	
	
	static Tidy tidy = new Tidy(); // obtain a new Tidy instance
	static StringBuffer cdData1 = new StringBuffer();
	static String cosmetic = "";
	static StringBuffer cdData2 = new StringBuffer();
	static String mathCalc = "";
	static String permission = "";
	static String missingPhoto = "";
	static HashMap testFaults = new HashMap();
	static HashMap realFaults = new HashMap();
	static HashMap names_to_id = new HashMap();
	static HashMap faultSeverities = new HashMap();
	static HashMap autoLabels = new HashMap();

	public static void main(String[] args) 
	{
		try
		{
			String fileName = null;
			try{
				fileName = args[0];
			}
			catch (Exception e){System.err.println("missing input file!");}
			System.out.println("processing file "+fileName);
			//tidy.setXHTML(true); // set desired config options using tidy setters 

			//result file
			BufferedWriter result = new BufferedWriter(new FileWriter("automatic.csv",true));

			//surface features; some are initialized outside of main 
			mathCalc = "";
			String error404 = "";
			cosmetic = "";
			String language = "";
			String css = "";
			String codeDump = "";
			String codeError = "";
			String functionalDisplay = "";
			missingPhoto = "";
			String wrongPage = "";
			String missingInfo = "";
			String authentication = "";
			permission = "";
			String session = "";
			String search = "";
			String database = "";
			String upload = "";

			//generates mapping between two files
			HashMap mapping = moveFiles("TRUE");

			//converts two input files into DOMs
			FileInputStream inputStreamCorrect = new FileInputStream("fileCTRUE.html");
			DOMDocumentImpl domCorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamCorrect, null); // run tidy, providing an input and output stream
			FileInputStream inputStreamIncorrect = new FileInputStream("fileITRUE.html");
			DOMDocumentImpl domIncorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamIncorrect, null); // run tidy, providing an input and output stream
			
			//WSOD
			if(domIncorrect == null || !domIncorrect.hasChildNodes()){
				wrongPage = "on";
				missingInfo = "on";
				//System.out.println("shortcut");
			}
			else if (equalTree("fileCTRUE.html","fileITRUE.html"))
			{
				//System.out.println("equalTree");
				wrongPage = "on";
				codeError = "on";
			}
			else{
				//wrong page
				wrongPage = C_wrongPage();
				//missing photo
				missingPhoto = C_missingPhoto(domCorrect, domIncorrect, mapping);
				//permission
				permission = C_text(domCorrect, domIncorrect, "permission");
				//session
				session = C_text(domCorrect, domIncorrect, "session");
				//missing info
				missingInfo = C_missingInfo();
				//code error
				codeError = C_text(domCorrect, domIncorrect, "codeError");
				//cosmetic
				/******************* cosmetic ************************/
				//math calc
				mathCalc = C_text(domCorrect, domIncorrect, "mathCalc");
				//functional display
				functionalDisplay = C_functionalDisplay();
				//database
				database = C_text(domCorrect, domIncorrect, "database");
				//authentication
				authentication = C_text(domCorrect, domIncorrect, "authentication");
				//error 404
				error404 = C_text(domCorrect, domIncorrect, "error404");
				//search
				search = C_text(domCorrect, domIncorrect, "search");
				//upload
				upload =  C_text(domCorrect, domIncorrect, "upload");
				//css
				css = C_css(domCorrect, domIncorrect, mapping);
				//language
				//code dump 
				codeDump = C_text(domCorrect, domIncorrect, "codeDump");
				if (permission.length() == 0 && authentication.length() == 0 && session.length() == 0)
					language = C_language(domCorrect, domIncorrect, mapping);
				else
					cosmetic = "";
				if (error404.length() != 0)
					cosmetic = "";
				if (functionalDisplay.length() == 0 && missingInfo.length() == 0 && codeError.length() == 0 && codeDump.length() == 0)
					cosmetic = "";

				int count = 0;
				if (wrongPage.length() > 0)
					count++;
				if (missingPhoto.length() > 0)
					count++;
				if (permission.length() > 0)
					count++;
				if (session.length() > 0)
					count++;
				if (missingInfo.length() > 0)
					count++;
				if (codeError.length() > 0)
					count++;
				if (mathCalc.length() > 0)
					count++;
				if (functionalDisplay.length() > 0)
					count++;	
				if (database.length() > 0)
					count++;		
				if (authentication.length() > 0)
					count++;		
				if (error404.length() > 0)
					count++;	
				if (search.length() > 0)
					count++;
				if (upload.length() > 0)
					count++;
				if (css.length() > 0)
					count++;

				if (count >= 4)
					cosmetic = "";
				
			}

			mathCalc = "";

			//process each input file, find auto_labels and predicted severity
			double score = -1;
				if (wrongPage.compareTo("on") == 0 && missingInfo.compareTo("on") == 0)
				score = 2.5;
			else if(permission.compareTo("on") == 0 && session.compareTo("on") != 0)
				score = 2.5;
			else if(missingPhoto.compareTo("on") == 0 && codeError.compareTo("on") != 0){
				if(cosmetic.compareTo("on") == 0)
					score = 0.5;
				else if (missingInfo.compareTo("on") == 0 && functionalDisplay.compareTo("on") == 0)
					score = 2;
				else
					score = 2.5;
			}
			else if(mathCalc.compareTo("on") == 0 && codeError.compareTo("on") == 0)
				score = 2.5;
			else if(functionalDisplay.compareTo("on") == 0 && codeError.compareTo("on") == 0)
				score = 2.5;
			else if(database.compareTo("on") == 0)
				score = 2.5;
			else if(codeError.compareTo("on") == 0)
				score = 2.5;
			else if(codeDump.compareTo("on") == 0){
				score = 2.5;
			}
			else if(authentication.compareTo("on") == 0)
				if(permission.compareTo("on") == 0)
					score = 3;
				else
					score = 2.5;
			else if(error404.compareTo("on") == 0)
				score = 2.5;
			else if(wrongPage.compareTo("on") == 0)
				score = 2;
			else if(session.compareTo("on") == 0)	//also 2 if permission is there
				score = 2;
			else if(search.compareTo("on") == 0 && cosmetic.compareTo("on") != 0)
				score = 2;
			else if(missingInfo.compareTo("on") == 0 && cosmetic.compareTo("on") != 0)
				score = 2;
			else if(upload.compareTo("on") == 0)
				score = 1.5;
			else if(missingPhoto.compareTo("on") == 0 && codeError.compareTo("on") == 0)
				score = 1.5;
			else if(mathCalc.compareTo("on") == 0 && cosmetic.compareTo("on") != 0)
				score = 1.5;
			else if(functionalDisplay.compareTo("on") == 0 && cosmetic.compareTo("on") != 0)
				score = 1.5;
			else if(css.compareTo("on") == 0)
				score = 1.5;
			else if(mathCalc.compareTo("on") == 0 && cosmetic.compareTo("on") == 0)
				score = 1.5;
			else if(language.compareTo("on") == 0){
				score = 1;
			}
			else if(functionalDisplay.compareTo("on") == 0 && cosmetic.compareTo("on") == 0)
				score = 1;
			else if(database.compareTo("on") == 0 && cosmetic.compareTo("on") == 0)
				score = 0.5;
			else if(missingInfo.compareTo("on") == 0 && cosmetic.compareTo("on") == 0)
				score = 0.75;
			else if(search.compareTo("on") == 0 && cosmetic.compareTo("on") == 0)
				score = 0.5;
			else{	//if no fault was seen
				score = 0.5;

			}
			
			System.out.println("wrongPage: " + wrongPage);
			System.out.println("missingPhoto: " + missingPhoto);
			System.out.println("permission: " + permission);
			System.out.println("session: " + session);
			System.out.println("missingInfo: " + missingInfo);
			System.out.println("codeError: " + codeError);
			System.out.println("mathCalc: " + mathCalc);
			System.out.println("functionalDisplay: " + functionalDisplay);
			System.out.println("database: " + database);
			System.out.println("authentication: " + authentication);
			System.out.println("error404: " + error404);
			System.out.println("search: " + search);
			System.out.println("cosmetic: " + cosmetic);
			System.out.println("language: " + language);
			System.out.println("codeDump: " + codeDump);
			System.out.println("upload: " + upload);
			System.out.println("css: " + css);
			System.out.println("session: " + session);

			if (cosmetic.compareTo("on") == 0 && score > 0.5 && !(codeDump.length() != 0 && mathCalc.length() != 0))
				score = 1;	
			
			System.out.println("SCORE: " + score);

			result.write("filename,SEVERITY,wrongPage,missingPhoto,permission,session,missingInfo,codeError,mathCalc,functionalDisplay,database,authentication,error404,search,cosmetic,language,codeDump,upload,css,session,\n");
			result.write(","+fileName+","+score+","+wrongPage+","+missingPhoto+","+permission+","+session+","+missingInfo+","+codeError+","+mathCalc+","+functionalDisplay+","+database+","+authentication+","+error404+","+search+","+cosmetic+","+language+","+codeDump+","+upload+","+css+","+session+",\n");
			result.close();		
		}
		catch (Exception e){e.printStackTrace();System.exit(0);}
	}

	static private Vector getSharedNodes(HashMap mapping, DOMNodeImpl dom){
		//get the nodes that are shared
		Vector typesShared = new Vector();
		TreeSet keys = new TreeSet(mapping.keySet());
		Iterator i = keys.iterator();
		while (i.hasNext())
		{
			Integer key = (Integer)i.next();
			DOMNodeImpl node = getNodeByID(dom,key.intValue());
			if (node != null)
				typesShared.add(node);
		}
		return typesShared;
	}

	//determines if attributes of nodes are the same
	static private boolean matchAttr(DOMNodeImpl n1, DOMNodeImpl n2){
		DOMAttrMapImpl attrs1 = (DOMAttrMapImpl)n1.getAttributes();
		DOMAttrMapImpl attrs2 = (DOMAttrMapImpl)n2.getAttributes();

		if (attrs1.getLength() != attrs2.getLength())
			return false;
	
		boolean found = false;
		for (int i = 0; i < attrs1.getLength(); i++)
		{
			DOMNodeImpl attr1 = (DOMNodeImpl)attrs1.item(i);
			for (int j = 0; j < attrs2.getLength(); j++)
			{
				DOMNodeImpl attr2 = (DOMNodeImpl)attrs2.item(i);
				try
				{
					if(attr1.getNodeValue().compareTo( ((DOMNodeImpl)attrs2.getNamedItem(attr1.getNodeName())).getNodeValue()) == 0)
						found = true;							
				}
				catch (NullPointerException e)
				{ return false;}
			}
			if (!found)
				return false;
			found = false;
		}
		return true;
	}

	//are the two trees the same
	static private boolean equalTree(String n1, String n2){
		BufferedReader buf1 = null;
		BufferedReader buf2 = null;	
		try
		{
			buf1 = new BufferedReader(new FileReader(n1));
			buf2 = new BufferedReader(new FileReader(n2));			
			String s1 = buf1.readLine();
			String s2 = buf2.readLine();
			
			while(s1 != null || s2 != null){
				//ignore the special handling of custom-inserted attributes
				s1 = s1.replaceAll("kinga_id=\"\\d+\"","");
				s2 = s2.replaceAll("kinga_id=\"\\d+\"","");				
				s1 = s1.replaceAll("kinga_id.\\d+.","");
				s2 = s2.replaceAll("kinga_id.\\d+.","");
				if (s1.compareTo(s2) != 0){
					buf1.close();
					buf2.close();
					return false;
				}
				s1 = buf1.readLine();
				s2 = buf2.readLine();
			}
		}
		catch (NullPointerException e)	{
			try
			{
				if(buf1 != null)
					buf1.close();
				if(buf2 != null)
					buf2.close();
				return false;
			}
			catch (Exception e1){e1.printStackTrace(); System.exit(0);}
		}
		catch (Exception e){e.printStackTrace(); System.exit(0);}

		return true;
	}

	static private boolean equalNode(DOMNodeImpl n1, DOMNodeImpl n2){
	
		if ( n1 != null && n2 != null && n1.getNodeName() != null && n2.getNodeName() != null && n1.getNodeName().compareTo(n2.getNodeName()) == 0)
		{
			if (n1.getNodeName().compareTo("#text") == 0 && n1.getNodeValue() != null && n2.getNodeValue() != null && n1.getNodeValue().compareTo(n2.getNodeValue()) == 0)
				return true;
			else if (n1.getNodeName().compareTo("#text") != 0)
			{
				DOMNodeListImpl children1 = (DOMNodeListImpl)n1.getChildNodes();
				DOMNodeListImpl children2 = (DOMNodeListImpl)n2.getChildNodes();
				if(children1.getLength() == children2.getLength() && matchAttr(n1,n2))
					return true;
			}
		}
		return false;
	}

	//find the differences between two trees
	static private Vector getUnmappedNodes(HashMap mapping, DOMNodeImpl dom){
		Vector allNodes = new Vector();
		allNodes = getAllNodes(dom,allNodes);
		Iterator i = allNodes.iterator();

		boolean found = false;
		Vector unsharedNodes = new Vector();

		Integer kinga_id = new Integer(-1);
		while (i.hasNext())
		{
			//if the allNodes member is not in the mapping, it is then unshared
			DOMNodeImpl n1 = (DOMNodeImpl)i.next();

			if (n1.getNodeName().compareTo("#text") == 0)
			{
				String text = n1.getNodeValue();
				int start = text.indexOf("kinga_id[");
				int end = text.indexOf("]",start);
				if (start != -1 && end != -1)
					kinga_id = new Integer(text.substring(start+9,end));
			}
			else {
				DOMAttrMapImpl attr = (DOMAttrMapImpl)n1.getAttributes();
				DOMNodeImpl id = (DOMNodeImpl)attr.getNamedItem("kinga_id");
				if (id != null && id.getNodeValue() != null)
					kinga_id = new Integer(id.getNodeValue());
			}
			if (!mapping.containsKey(kinga_id) && n1.getNodeName().trim().length() != 0 && n1.getNodeName().compareTo("br") != 0)
				unsharedNodes.add(n1);
		}

		return unsharedNodes;
	}

	//formats the files to be used with this code; some bookkeeping code was inserted earlier by parseBothPaper3.py
	//the input to the function is either TRUE or FALSE which determines whether or not that attribute values
	//themselves are considered when checking if two elements are the same
	static private HashMap moveFiles(String TF){
		try
		{
			HashMap mapping = new HashMap();
			String s = null;
			int count = 0;
			BufferedReader parsed = new BufferedReader(new FileReader("parsed"+TF+".html"));
			BufferedWriter fileC = new BufferedWriter(new FileWriter("fileC"+TF+".html"));
			BufferedWriter fileI = new BufferedWriter(new FileWriter("fileI"+TF+".html"));
			StringBuffer html = new StringBuffer();
			while ((s = parsed.readLine()) != null)
				html.append(s+"\n");
			int img = html.indexOf("<img");
			while(img != -1){
				int end = html.indexOf(">",img);
				if (html.charAt(end-1) == '/')
					html.setCharAt(end-1,' ');
				img = html.indexOf("<img",img+1);

			}
			img = html.indexOf("<IMG");
			while(img != -1){
				int end = html.indexOf(">",img+1);
				if (html.charAt(end-1) == '/')
					html.setCharAt(end-1,' ');
				img = html.indexOf("<IMG",img);
			}
			BufferedWriter htmlOut = new BufferedWriter(new FileWriter("parsed"+TF+".html"));
			int index = 0;
			while (index < html.length()){
				if (index + 255 < html.length())
					htmlOut.write(html.substring(index,index+255));
				else
					htmlOut.write(html.substring(index,html.length()));
				index = index + 255;
			}
			htmlOut.close();
			parsed = new BufferedReader(new FileReader("parsed"+TF+".html"));
			while ((s = parsed.readLine()) != null)
			{
				if (s.trim().compareTo("###################") == 0)
					count++;
				if (count == 0)
					fileC.write(s+"\n");
				else if (count == 1 && s.compareTo("###################") != 0)
					fileI.write(s+"\n");
				else{
					int comma = s.indexOf(",");
					if (comma != -1)
					{
						Integer num1 = new Integer(s.substring(1,comma));
						Integer num2 = new Integer(s.substring(comma+1,s.length()-1));
						mapping.put(num1,num2);
						mapping.put(num2,num1);
					}
				}
			}
			parsed.close();
			fileC.close();
			fileI.close();

			return mapping;			
		}
		catch (Exception e)	{e.printStackTrace(); System.exit(0);}
		return null;
	}

	//****************************
	// Get minimum of three values
	//****************************
	
	static private int Minimum (int a, int b, int c) {
		int mi;
		mi = a;
		if (b < mi) 
			mi = b;
		if (c < mi) 
			mi = c;
		return mi;
	}

	//*****************************
	// Compute Levenshtein distance
	//*****************************
	static public int LD (String s, String t) throws OutOfMemoryError{
		  int d[][]; // matrix
		  int n; // length of s
		  int m; // length of t
		  int i; // iterates through s
		  int j; // iterates through t
		  char s_i; // ith character of s
		  char t_j; // jth character of t
		  int cost; // cost

		  // Step 1
		  n = s.length ();
		  m = t.length ();
		  if (n == 0) 
			  return m;
		  if (m == 0) 
			  return n;
		  d = new int[n+1][m+1];
		  
		  // Step 2
		  for (i = 0; i <= n; i++) 
			  d[i][0] = i;
		  for (j = 0; j <= m; j++) 
			  d[0][j] = j;
		  
		  // Step 3
		  for (i = 1; i <= n; i++) {
			  s_i = s.charAt (i - 1);
	      
			  // Step 4
			  for (j = 1; j <= m; j++) {
				  t_j = t.charAt (j - 1);
		
				  // Step 5
				  if (s_i == t_j) 
					  cost = 0;
				  else 
					  cost = 1;
		
				  // Step 6
				  d[i][j] = Minimum (d[i-1][j]+1, d[i][j-1]+1, d[i-1][j-1] + cost);
			  }
		  }
		  
		  // Step 7
		  return d[n][m];
	  }

	//was it a language error
	static private String C_language(DOMNodeImpl domCorrect, DOMNodeImpl domIncorrect, HashMap mapping){
		Vector v1 = new Vector();
		v1 = DOM2Text(domCorrect, v1);
		Vector v2 = new Vector();
		v2 = DOM2Text(domIncorrect,v2);

		if (v1.size() == v2.size())
		{
			for (int i = 0; i < v1.size() ; i++)
			{
				String s1 = strip(v1.get(i).toString());
				String s2 = strip(v2.get(i).toString());
				String originalUnmatched = "";
				String newUnmatched = "";
				
				if (s1.length() + 1 == s2.length() && s2.indexOf("\\") != -1){
					return "on";
				}
				try
				{
					int levenstein = LD(s1,s2);
					if (levenstein > 0 && levenstein <= 10 && (s2.indexOf("&") != -1 || levenstein == 2)){
						return "on";		
					}
				}
				catch (OutOfMemoryError e){}
			}
		}

		return "";
	}

	//was it the wrong page
	static private 	String C_wrongPage(){
		try
		{
			HashMap mapping = moveFiles("FALSE");

			FileInputStream inputStreamCorrect = new FileInputStream("fileCFALSE.html");
			DOMDocumentImpl domCorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamCorrect, System.out); // run tidy, providing an input and output stream
			FileInputStream inputStreamIncorrect = new FileInputStream("fileIFALSE.html");
			DOMDocumentImpl domIncorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamIncorrect, System.out); // run tidy, providing an input and output stream

			//wrong page is different from an error because there will still be some html
			Vector shared = getSharedNodes(mapping, domCorrect);
			Iterator i = shared.iterator();
			Vector other = new Vector();
			boolean html = false;
			boolean head = false;
			boolean body = false;
			while (i.hasNext())
			{
				DOMNodeImpl node = (DOMNodeImpl)i.next();
				if (node != null && node.getNodeName() != null)
				{
					//System.out.println("node.getNodeName()" + node.getNodeName());
					if(node.getNodeName().compareTo("html") == 0)
						html = true;
					else if(node.getNodeName().compareTo("head") == 0)
						head = true;
					else if(node.getNodeName().compareTo("body") == 0)
						body = true;
				}
			}

			//should always have these nodes for html
			if (html && head && body)
			{
				other = getAllNodes(domIncorrect,other);
				float ratio = ((float)(((float)shared.size()/2) / other.size())); 
				if ( ratio <= 0.25)//.34,.28
					return "on";
			}

			return "";			
		}
		catch (Exception e){e.printStackTrace();System.exit(0);}
		return null;
	}

	//was it a css error
	static private String C_css(DOMNodeImpl domCorrect, DOMNodeImpl domIncorrect, HashMap mapping){

		Vector unmappedCorrect = getUnmappedNodes(mapping, domCorrect);
		Vector unmappedIncorrect = getUnmappedNodes(mapping, domIncorrect);
		
		Vector linksOld = getNodeFromVector(unmappedCorrect, "link");
		Vector linksNew = getNodeFromVector(unmappedIncorrect, "link");

		//if the stylesheet is missing
		if (linksOld.size() != linksNew.size())
			return "on";
		
		//if the stylesheet has changed
		else if (linksOld.size() == 1 && linksNew.size() == 1)
		{
			DOMNodeImpl liOld = (DOMNodeImpl)linksOld.get(0);
			DOMNodeImpl liNew = (DOMNodeImpl)linksNew.get(0);
			DOMAttrMapImpl attrs1 = (DOMAttrMapImpl)liOld.getAttributes();
			DOMNodeImpl src1 = (DOMNodeImpl)attrs1.getNamedItem("href"); 
			DOMAttrMapImpl attrs2 = (DOMAttrMapImpl)liNew.getAttributes();
			DOMNodeImpl src2 = (DOMNodeImpl)attrs2.getNamedItem("href");
			if (src1.getNodeValue().compareTo(src2.getNodeValue()) != 0)
				return "on";
		}
		return "";
	}

	//missing image
	static private String C_missingPhoto(DOMNodeImpl domCorrect, DOMNodeImpl domIncorrect, HashMap mapping){

		Vector unmappedCorrect = getUnmappedNodes(mapping, domCorrect);
		Vector unmappedIncorrect = getUnmappedNodes(mapping, domIncorrect);
		
		Vector imagesOld = getNodeFromVector(unmappedCorrect, "img");
		Vector imagesNew = getNodeFromVector(unmappedIncorrect, "img");

		//entire image tag is missing - should only be 1 image, otherwise more serious error
		if (imagesOld.size() != imagesNew.size())
			return "on";
		//only one src link is broken - otherwise more serious error
		else if (imagesOld.size() <= 10) //may want to make this more sophisticated - check for similarities?
		{
			for (int i = 0; i < imagesOld.size(); i++ )
			{
				DOMNodeImpl imgOld = (DOMNodeImpl)imagesOld.get(i);
				DOMNodeImpl imgNew = (DOMNodeImpl)imagesNew.get(i);
				DOMAttrMapImpl attrs1 = (DOMAttrMapImpl)imgOld.getAttributes();
				DOMNodeImpl src1 = (DOMNodeImpl)attrs1.getNamedItem("src"); 
				DOMAttrMapImpl attrs2 = (DOMAttrMapImpl)imgNew.getAttributes();
				DOMNodeImpl src2 = (DOMNodeImpl)attrs2.getNamedItem("src");
				if (src1.getNodeValue().compareTo(src2.getNodeValue()) != 0)
					return "on";
				}
		}
		return "";
	}

	//something wrong with the form elements
	static private String C_functionalDisplay(){
		try
		{
			HashMap mapping = moveFiles("FALSE");

			FileInputStream inputStreamCorrect = new FileInputStream("fileCFALSE.html");
			DOMDocumentImpl domCorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamCorrect, System.out); // run tidy, providing an input and output stream
			FileInputStream inputStreamIncorrect = new FileInputStream("fileIFALSE.html");
			DOMDocumentImpl domIncorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamIncorrect, System.out); // run tidy, providing an input and output stream

			Vector unmappedCorrect = getUnmappedNodes(mapping, domCorrect);
			Vector unmappedIncorrect = getUnmappedNodes(mapping, domIncorrect);
			
			Vector functionalOld = getNodeFromVector(unmappedCorrect, "form");
			Vector functionalNew = getNodeFromVector(unmappedIncorrect, "form");
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "input"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "input"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "applet"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "applet"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "button"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "button"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "option"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "option"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "optgroup"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "optgroup"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "iframe"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "iframe"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "script"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "script"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "select"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "select"));
			functionalOld.addAll(getNodeFromVector(unmappedCorrect, "a"));
			functionalNew.addAll(getNodeFromVector(unmappedIncorrect, "a"));

			if (functionalOld.size() < functionalNew.size())
				permission = "on";	//this is kind of annoying, make sure you need it
			if (functionalOld.size() > functionalNew.size())
				return "on";
			
			//make sure that this works
			Vector inputOld = new Vector();
			inputOld = getNodeByName(domCorrect, "input", inputOld);
			Vector inputNew = new Vector();
			inputNew = getNodeByName(domIncorrect, "input", inputNew);
			for (int i = 0; i < inputOld.size(); i++)
			{
				DOMNodeImpl node1 = (DOMNodeImpl)inputOld.get(i);
				DOMNodeImpl node2 = (DOMNodeImpl)inputNew.get(i);
				DOMAttrMapImpl attrs1 = (DOMAttrMapImpl)node1.getAttributes();
				DOMNodeImpl src1 = (DOMNodeImpl)attrs1.getNamedItem("type"); 
				DOMAttrMapImpl attrs2 = (DOMAttrMapImpl)node2.getAttributes();
				DOMNodeImpl src2 = (DOMNodeImpl)attrs2.getNamedItem("type");
				if (src1 != null && src1.getNodeValue() != null && src1.getNodeValue().compareTo(src2.getNodeValue()) != 0){
					return "on";
				}
				DOMNodeImpl node = src1;
				boolean txt = false;
				if (src1 != null && src1.getNodeValue() != null && src1.getNodeValue().compareTo("text") == 0)
					txt = true;
				src1 = (DOMNodeImpl)attrs1.getNamedItem("value"); 
				src2 = (DOMNodeImpl)attrs2.getNamedItem("value");
				if (src1 != null && src2 != null)
				{
					String num1 = getNumber(src1.getNodeValue());
					String num2 = getNumber(src2.getNodeValue());
					if (txt && num1.length() > 0 && num1.compareTo(num2) != 0){
						mathCalc = "on";
					}
					DOMNodeImpl type = (DOMNodeImpl)attrs1.getNamedItem("type"); 
					if (type != null && type.getNodeValue() != null && type.getNodeValue().compareTo("checkbox") != 0 &&  src1.getNodeValue() != null && src1.getNodeValue().compareTo(src2.getNodeValue()) != 0){
						return "on";
					}
				}
			}

			inputOld = new Vector();
			inputOld = getNodeByName(domCorrect, "a", inputOld);
			inputNew = new Vector();
			inputNew = getNodeByName(domIncorrect, "a", inputNew);
			for (int i = 0; i < inputOld.size(); i++)
			{
				DOMNodeImpl node1 = (DOMNodeImpl)inputOld.get(i);
				DOMNodeImpl node2 = (DOMNodeImpl)inputNew.get(i);
				DOMAttrMapImpl attrs1 = (DOMAttrMapImpl)node1.getAttributes();
				DOMNodeImpl src1 = (DOMNodeImpl)attrs1.getNamedItem("href"); 
				DOMAttrMapImpl attrs2 = (DOMAttrMapImpl)node2.getAttributes();
				DOMNodeImpl src2 = (DOMNodeImpl)attrs2.getNamedItem("href");
				if (src1 !=null && src2 != null && src1.getNodeValue() != null && src1.getNodeValue().compareTo(src2.getNodeValue()) != 0){
					return "";	//took out because will produce more FP than FN
				}
			}

			return "";
		}
		catch (Exception e){e.printStackTrace(); System.exit(0);}
		return "";
	}

	static private void printNodeVector(Vector v, String name){
		for (int i = 0; i < v.size(); i++)
		{
			DOMNodeImpl node = (DOMNodeImpl)v.get(i);
			if (node.getNodeName().compareTo("#text") != 0)
				System.out.println("v["+i+"] = " + node.getNodeName());
			else 
				System.out.println("v["+i+"] = " + node.getNodeValue());
		}
	}

	//if there is missing information
	static private String C_missingInfo(){
		try{
			HashMap mapping = moveFiles("FALSE");

			FileInputStream inputStreamCorrect = new FileInputStream("fileCFALSE.html");
			DOMDocumentImpl domCorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamCorrect, null); // run tidy, providing an input and output stream
			FileInputStream inputStreamIncorrect = new FileInputStream("fileIFALSE.html");
			DOMDocumentImpl domIncorrect = (DOMDocumentImpl)tidy.parseDOM(inputStreamIncorrect, null); // run tidy, providing an input and output stream

			Vector unmappedCorrect = getUnmappedNodes(mapping, domCorrect);
			Vector unmappedIncorrect = getUnmappedNodes(mapping, domIncorrect);

			printNodeVector(unmappedCorrect,"unmappedCorrect");
			printNodeVector(unmappedIncorrect,"unmappedIncorrect");

			Vector allIncorrect = new Vector();
			allIncorrect = getAllNodes(domIncorrect, allIncorrect);
			printNodeVector(allIncorrect,"allIncorrect");
			
			//uses heuristics to determine if missing information was significant
			if( ( ((float)unmappedIncorrect.size()) /((float)allIncorrect.size()) ) <= 0.1 && missingPhoto.length() == 0)//may want to change to compare between unmatched of each file
																	//try if unmapped = #document, and >=15 items, not cosm
																	//maybe remove missingPhoto																//if too many things checked, turn off
			{
					cosmetic = "on";	
			}
			if ( ((float)unmappedCorrect.size()) >= 15 && ((float)unmappedIncorrect.size()) <= 2 )
				cosmetic = "";	

			if (unmappedCorrect.size() > unmappedIncorrect.size())
				return "on";

			return "";			
		}
		catch (Exception e){e.printStackTrace();System.exit(0);}
		return null;

	}

	static private Vector DOM2Text(DOMNodeImpl node, Vector v){
		//System.out.println("node name: " + node.getNodeName());
		if (node.getNodeName().compareTo("#text") == 0){
			if (node.getParentNode() != null && node.getParentNode().getNodeName().compareTo("style") != 0)
			{
				String val = node.getNodeValue().toLowerCase();
				v.add(val+"\n");
				//System.out.println("Added txt: " + node.getNodeValue());
				//System.out.println("parent: " + node.getParentNode().getNodeName());
			}
		}
		else {
			DOMNodeListImpl children = (DOMNodeListImpl)node.getChildNodes();
			for (int i = 0; i < children.getLength(); i++){
				//System.out.println("child " + i);
				v = DOM2Text((DOMNodeImpl)children.item(i),v);
			}
		}

		return v;		
	}

	static private String strip(String s){
		StringBuffer clean = new StringBuffer();
		while (s.lastIndexOf("kinga_id") != -1)
		{
			int index = s.indexOf("kinga_id");
			int end = s.indexOf("]",index);
			clean.append(s.substring(0,index));
			s = s.substring(end+1,s.length());
		}
		if (clean.toString().trim().length() == 0)
			return s;
		else
			return clean.toString();

	}

	static private String diff (String s1, String s2, String which){
	
		s1 = strip(s1);
		s2 = strip(s2);
		//System.out.println(s1);
		//System.out.println(s2);
		String[] x = s1.split("\\n");
		String[] y = s2.split("\\n");//   \\n
	
		// number of lines of each file
		int M = x.length;
		int N = y.length;
	
		// opt[i][j] = length of LCS of x[i..M] and y[j..N]
		int[][] opt = new int[M+1][N+1];
	
		// compute length of LCS and all subproblems via dynamic programming
		for (int i = M-1; i >= 0; i--) {
		    for (int j = N-1; j >= 0; j--) {
			if (x[i].equals(y[j]))
			    opt[i][j] = opt[i+1][j+1] + 1;
			else 
			    opt[i][j] = Math.max(opt[i+1][j], opt[i][j+1]);
		    }
		}
	
		// recover LCS itself and print out non-matching lines to standard output
		StringBuffer retX = new StringBuffer();
		StringBuffer retY = new StringBuffer();
		int i = 0, j = 0;
		while(i < M && j < N) {
		    if (x[i].equals(y[j])) {
			i++;
			j++;
		    }
		    else if (opt[i+1][j] >= opt[i][j+1]) retX.append(x[i++]+"\n");
		    else                                 retY.append(y[j++]+"\n");
		}

		// dump out one remainder of one string if the other is exhausted
		while(i < M || j < N) {
		    if      (i == M) retY.append(y[j++]+"\n");
		    else if (j == N) retX.append(x[i++]+"\n");
		}
		if(which.compareTo("first") == 0)
			return retX.toString();
		else
			return retY.toString();
	}

	static private int count(StringBuffer string, String sub){
		StringBuffer temp = new StringBuffer(string);
		int index = -1;
		int count = 0;
		index = temp.lastIndexOf(sub);
		while (index != -1)
		{
			count++;
			//string.delete(0,index);
			temp = new StringBuffer(temp.delete(index,temp.length()));
			//System.out.println("count:");
			index = temp.lastIndexOf(sub);
		}
		return count;
	}

	static private StringBuffer removeCDATA(StringBuffer string, int num){
		int end = string.lastIndexOf("]]");
		int start = string.lastIndexOf("[CDATA[");
		if (start == -1)
			start = string.lastIndexOf("[cdata[");
		while (start != -1 && end > start)
		{
			if(num == 1)
				cdData1.append(string.substring(start,end+2) + "\n");
			else
				cdData2.append(string.substring(start,end+2) + "\n");
			string = new StringBuffer(string.delete(start,end+2));
			end = string.lastIndexOf("]]");
			start = string.lastIndexOf("[CDATA[");
			if (start == -1)
				start = string.lastIndexOf("[cdata[");
		}

		start = string.indexOf("function");
		end = string.indexOf("}",start);
		while (start != -1 && end > start)
		{
			if(num == 1)
				cdData1.append(string.substring(start,end+2) + "\n");
			else
				cdData2.append(string.substring(start,end+2) + "\n");
			string = new StringBuffer(string.delete(start,end+1));
			start = string.indexOf("function");
			end = string.indexOf("}",start);
		}

		return string;
	}

	static private StringBuffer removeKingaID(StringBuffer string){
		int start = string.lastIndexOf("kinga_id[");
		int end = string.indexOf("]",start);
		while (start != -1)
		{
			string = new StringBuffer(string.delete(start,end+1));
			start = string.lastIndexOf("kinga_id[");
			end = string.indexOf("]",start);		}

		return string;
	}

	//checks for natural language text in the outputs specified by the function passed in
	static private String C_text(DOMNodeImpl correct, DOMNodeImpl incorrect, String function){
		Vector v1 = new Vector();
		v1 = DOM2Text(correct, v1);

		Vector v2 = new Vector();
		v2 = DOM2Text(incorrect,v2);

		StringBuffer text1 = new StringBuffer();
		StringBuffer text2 = new StringBuffer();
		for (int i = 0; i < v1.size(); i++)
			text1.append(v1.get(i).toString());
		for (int i = 0; i < v2.size(); i++)
			text2.append(v2.get(i).toString());
		text1 = removeCDATA(text1,1);
		text2 = removeCDATA(text2,2);
		text1 = removeKingaID(text1);
		text2 = removeKingaID(text2);

		HashMap mapping = moveFiles("TRUE");

		Vector unmappedCorrect = getUnmappedNodes(mapping, correct);
		Vector unmappedIncorrect = getUnmappedNodes(mapping, incorrect);

		Vector allCorrect = new Vector();
		Vector allIncorrect = new Vector();
		allCorrect = getAllNodes(correct,allCorrect);
		allIncorrect = getAllNodes(incorrect,allIncorrect);

		if (function.compareTo("permission") == 0)
		{
			if (text2.lastIndexOf("permission") != -1 && count(text1,"permission") != count(text2,"permission"))
				return "on";
			if (text2.lastIndexOf("do not have") != -1 && count(text1,"do not have") < count(text2,"do not have"))
				return "on";
			if (text2.lastIndexOf("don't have") != -1 && count(text1,"don't have") < count(text2,"don't have"))
				return "on";
			if (text2.lastIndexOf("lock") != -1 && count(text1,"lock") < count(text2,"lock"))
				return "on";
			if (text2.lastIndexOf("access") != -1 && count(text1,"access") < count(text2,"access"))
				return "on";
			if (text2.lastIndexOf("not allowed") != -1 && count(text1,"not allowed") < count(text2,"not allowed"))
				return "on";
			if (text2.lastIndexOf("read") != -1 && count(text1,"read") < count(text2,"read"))
				return "on";
			//if (text2.lastIndexOf("edit") != -1 && count(text1,"edit") < count(text2,"edit"))
			//	return "on";
			//if (text2.lastIndexOf("admin") != -1 && count(text1,"admin") < count(text2,"admin"))
			//	return "on";
			if (text2.lastIndexOf("none") != -1 && count(text1,"none") < count(text2,"none"))
				return "on";
			if (text2.lastIndexOf("update") != -1 && count(text1,"update") < count(text2,"update"))
				return "on";
			if (text2.lastIndexOf("banned") != -1 && count(text1,"banned") < count(text2,"banned"))
				return "on";
		}
		else if (function.compareTo("session") == 0)
		{
			if (text2.lastIndexOf("expired") != -1 && count(text1,"expired") < count(text2,"expired"))
				return "on";
			if (text2.lastIndexOf("timed out") != -1 && count(text1,"timed out") < count(text2,"timed out"))
				return "on";
			if (text2.lastIndexOf("log in") != -1 && count(text1,"log in") < count(text2,"log in"))
				return "on";
			if (text2.lastIndexOf("login") != -1 && count(text1,"login") < count(text2,"login"))
				return "on";
			if (text2.lastIndexOf("token") != -1 && count(text1,"token") < count(text2,"token"))
				return "on";
		}
		else if (function.compareTo("codeError") == 0)
		{
			if( ( ((float)unmappedIncorrect.size()) /((float)allIncorrect.size()) ) <= 0.1 && missingPhoto.length() == 0){
				cosmetic = "on";
			}
			//if ( ((float)unmappedCorrect.size()) >= 15 && ((float)unmappedIncorrect.size()) <= 2 )
			//{cosmetic = "";	System.out.println("in!");}
			//if (cdData1.toString().compareTo(cdData2.toString()) != 0)
			//	return "on";
			if (text2.lastIndexOf("error") != -1 && count(text1,"error") < count(text2,"error"))
				return "on";
			if (text2.lastIndexOf("incorrect") != -1 && count(text1,"incorrect") < count(text2,"incorrect"))
				return "on";
			if (text2.lastIndexOf("not") != -1 && count(text1,"not") < count(text2,"not") && text2.lastIndexOf("upload") == -1)
				return "on";
			if (text2.lastIndexOf("unable") != -1 && count(text1,"unable") < count(text2,"unable"))
				return "on";
			if (text2.lastIndexOf("unexpected") != -1 && count(text1,"unexpected") < count(text2,"unexpected"))
				return "on";
			if (text2.lastIndexOf("exception") != -1 && count(text1,"exception") < count(text2,"exception"))
				return "on";
			if (text2.lastIndexOf("server") != -1 && count(text1,"server") < count(text2,"server"))
				return "on";
		}
		else if (function.compareTo("mathCalc") == 0)
		{
			if (text2.lastIndexOf("miscalculation") != -1 && count(text1,"miscalculation") < count(text2,"miscalculation"))
				return "on";
			String number1 = getNumber(diff(text1.toString(),text2.toString(), "first"));
			String number2 = getNumber(diff(text1.toString(),text2.toString(), "second"));

			if (number1.compareTo(number2) != 0)
				return "on";
		}
		else if (function.compareTo("database") == 0)
		{
			if( ( ((float)unmappedIncorrect.size()) /((float)allIncorrect.size()) ) <= 0.1 && missingPhoto.length() == 0)
			{cosmetic = "on"; }
			//if ( ((float)unmappedCorrect.size()) >= 15 && ((float)unmappedIncorrect.size()) <= 2 )
			//	cosmetic = "";	
			if (text2.lastIndexOf("unknown") != -1 && count(text1,"unknown") < count(text2,"unknown"))
				return "on";
			if (text2.lastIndexOf("sql") != -1 && count(text1,"sql") < count(text2,"sql"))
				return "on";
			if (text2.lastIndexOf("database") != -1 && count(text1,"database") < count(text2,"database"))
				return "on";
			if (text2.lastIndexOf("query") != -1 && count(text1,"query") < count(text2,"query"))
				return "on";
			if (text2.lastIndexOf("select") != -1 && count(text1,"select") < count(text2,"select"))
				return "on";
			if (text2.lastIndexOf("update") != -1 && count(text1,"update") < count(text2,"update"))
				return "on";
		}
		else if (function.compareTo("authentication") == 0)
		{
			if (text2.lastIndexOf("password") != -1 && count(text1,"password") < count(text2,"password"))
				return "on";
			if (text2.lastIndexOf("username") != -1 && count(text1,"username") < count(text2,"username"))
				return "on";
			if (text2.lastIndexOf("login") != -1 && count(text1,"login") < count(text2,"login"))
				return "on";
		}
		else if (function.compareTo("error404") == 0)
		{
			if (text2.lastIndexOf("404") != -1 && count(text1,"404") < count(text2,"404"))
				return "on";
			if (text2.lastIndexOf("not found") != -1 && count(text1,"not found") < count(text2,"not found"))
				return "on";
		}
		else if (function.compareTo("search") == 0)
		{
			if (text2.lastIndexOf("results") != -1 )
				return "on";
		}
		else if (function.compareTo("codeDump") == 0)
		{
			if( ( ((float)unmappedIncorrect.size()) /((float)allIncorrect.size()) ) <= 0.1 && missingPhoto.length() == 0)
				{cosmetic = "on"; }
			//if ( ((float)unmappedCorrect.size()) >= 15 && ((float)unmappedIncorrect.size()) <= 2 )
			//	cosmetic = "";	
			if (text2.lastIndexOf("warning") != -1 && count(text1,"warning") < count(text2,"warning")){
				cosmetic = "on";
			}
			if (text2.lastIndexOf(" line") != -1 && count(text1," line") < count(text2," line"))
				return "on";
			if (text2.lastIndexOf(".php") != -1 && count(text1,".php") < count(text2,".php"))
				return "on";
			if (text2.lastIndexOf("boolean") != -1 && count(text1,"boolean") < count(text2,"boolean"))
				return "on";
			if (text2.lastIndexOf("{") != -1 && count(text1,"{") < count(text2,"{"))
				return "on";
			if (text2.lastIndexOf("=") != -1 && count(text1,"=") < count(text2,"="))
				return "on";
			if (text2.lastIndexOf("(") != -1 && count(text1,"(") < count(text2,"("))
				return "on";
		}
		else if (function.compareTo("upload") == 0)
		{
			if ((text2.lastIndexOf("failed") != -1 && count(text1,"failed") < count(text2,"failed")) &&
				(text2.lastIndexOf("upload") != -1 && count(text1,"upload") < count(text2,"upload")))
				return "on";
			if ((text2.lastIndexOf("not") != -1 && count(text1,"not") < count(text2,"not")) &&
				(text2.lastIndexOf("upload") != -1 && count(text1,"upload") < count(text2,"upload")))
				return "on";
			if (text2.lastIndexOf("no image") != -1 && count(text1,"no image") < count(text2,"no image"))
				return "on";
		}
		return "";
	}

	static private String getNumber(String s){
		StringBuffer buf = new StringBuffer();
		int found = 0;

		//may want to change so that it only adds floats
		for (int i = 0; i < s.length(); i++)
		{
			int c = (int)s.charAt(i);
			if (c >= 48 && c <= 57 || c == 46)
				if (c == 46)
					buf.append(".");
				else
					buf.append((new Integer(c-48)).toString());
		}
		return buf.toString();
	}

	static private Vector getNodeFromVector(Vector v, String name){
		Vector ret = new Vector();
		Iterator i = v.iterator();
		while (i.hasNext())
		{
			DOMNodeImpl node = (DOMNodeImpl)i.next();
			//System.out.println("node Name: " + node.getNodeName());
			if (name.compareTo(node.getNodeName()) == 0)
				ret.add(node);

		}
		return ret;
	}

	static private DOMNodeImpl getNodeByID(DOMNodeImpl node, int id){
		//System.out.println("node name: " + node.getNodeName());
		if (node.getNodeName().compareTo("#text") != 0){
			DOMAttrMapImpl attrs = (DOMAttrMapImpl)node.getAttributes();
			DOMNodeImpl identifier = (DOMNodeImpl)attrs.getNamedItem("kinga_id"); 
			if (identifier != null)
			{
				String value = identifier.getNodeValue();
				//System.out.println("non-text node KINGA_ID: " + value);
				Integer nodeID = new Integer(value);
				if (nodeID.intValue() == id)
					return node;
			}
			DOMNodeListImpl children = (DOMNodeListImpl)node.getChildNodes();
			for (int i = 0; i < children.getLength(); i++){
				DOMNodeImpl child = getNodeByID((DOMNodeImpl)children.item(i),id);
				if(child != null)
					return child;
			}
		}
		else{
			String value = node.getNodeValue();
			if (value.trim().length() > 0)
			{
				//System.out.println("Text node:" + value+":");
				int index = value.lastIndexOf("kinga_id[");
				try{
					Integer nodeID = new Integer(value.substring(index+9,value.lastIndexOf("]")));
					if (nodeID.intValue() == id)
						return node;
				}catch (Exception e){}
			}
		}
		//System.out.println("couldn't find index");
		return null;
	}

	static private Vector getNodeByName(DOMNodeImpl node, String name, Vector nodes){
		//System.out.println("node name: " + node.getNodeName());
		if (node.getNodeName().compareTo(name) == 0)
			nodes.add(node);

		DOMNodeListImpl children = (DOMNodeListImpl)node.getChildNodes();
		for (int i = 0; i < children.getLength(); i++)
			nodes = getNodeByName((DOMNodeImpl)children.item(i),name,nodes);

		return nodes;
	}

	static private Vector getAllNodes(DOMNodeImpl node, Vector nodes){
		//System.out.println("node name: " + node.getNodeName());
		nodes.add(node);

		DOMNodeListImpl children = (DOMNodeListImpl)node.getChildNodes();
		for (int i = 0; i < children.getLength(); i++)
			nodes = getAllNodes((DOMNodeImpl)children.item(i),nodes);

		return nodes;
	}

}
